package Apache::ZopeProxy;

use strict;
use vars qw(@ISA $VERSION);

use Apache::Constants qw(:response);
use Apache::File;
use LWP::UserAgent ();
use Time::HiRes qw(time);
use MD5;
use File::Find;

# the following path has to be writeable for the apache user
use constant DEFAULT_ZOPEPROXY_PATH => '/var/cache/zopeproxy';

@ISA = qw(LWP::UserAgent);
$VERSION = '1.00';

my $UA = __PACKAGE__->new;
$UA->agent(join '/', __PACKAGE__, $VERSION);

# global hash for registered cacheable objects
my %objects = ();
my %timeout = ();
my %noauth = ();
my %nocookie = ();
my %anon = ();

# reference to a global hash of allowed IPs for management requests
my $allowedip;

# get the signature of an object (based on the REQUEST)
sub getSignature {
   my ($base, $object, $parms, $r) = @_;
   my $signature = '/'.MD5->hexhash($parms);
   unless ($noauth{$base}->{$object}) {
      my $auth = $r->header_in('Authorization');
      $signature .= '.'.MD5->hexhash($auth) if $auth;
   }
   unless ($nocookie{$base}->{$object}) {
      my $cookie = $r->header_in('Cookie');
      $signature .= ':'.MD5->hexhash($cookie) if $cookie;
   }
   return $signature;
}

# fall thru to mod_proxy
sub proxyRequest {
   my ($base, $uri, $r, $reason) = @_;
   $r->log_reason($reason) if $reason;
   $r->proxyreq(1);
   $r->uri($base.$uri);
   $r->filename("proxy:$base$uri");
   $r->handler('proxy-server');
   return OK;
}

# initialize the base registration for URLs to be cached
sub initializeBase {
   my ($path, $base, $r) = @_;
   # this server is not initialized yet, load the registration data
   $objects{$base} = {};
   $timeout{$base} = {};
   $noauth{$base} = {};
   $nocookie{$base} = {};
   $anon{$base} = {};
   my $fh = Apache::File->new($path.'registered-urls');
   my $dump = Apache::File->new('>'.$path.'registered-urls-dump');
   if ($fh) {
      while (<$fh>) {
         chomp;
	 next if /^\s*$/;
	 my ($object, $timeout, $content_type, @attrs)  = split(/\t/);
	 my $hash = MD5->hexhash($object);
	 mkdir $path.$hash, 0755 if ! -d $path.$hash;
	 my @segments = split('/', $object);
	 $objects{$base}->{$object} = $content_type;
	 $timeout{$base}->{$object} = $timeout;
	 my %attrs = map { $_ => 1 } @attrs;
	 $noauth{$base}->{$object} = 1 if $attrs{noauth};
	 $nocookie{$base}->{$object} = 1 if $attrs{nocookie};
	 $anon{$base}->{$object} = 1 if $attrs{anon};
	 print $dump $object, "\n";
	 print $dump "\tpath=", $hash, "\n";
	 print $dump "\ttype=", $content_type, "\n";
	 print $dump "\tttl=", $timeout{$base}->{$object}, "\n";
	 print $dump "\tnoauth\n" if $noauth{$base}->{$object};
	 print $dump "\tnocookie\n" if $nocookie{$base}->{$object};
	 print $dump "\tanon\n" if $anon{$base}->{$object};
      }
   } else {
      $r->log_reason("Can't read ${path}registered-urls for $base");
   }
}

# if the request is for one of the configured objects, use the proxy handler
sub handler {
   my $r = shift;
   my $uri = $r->uri;
   if ($uri =~ /^\/manage_ZopeProxyRemoteAction\//) {
      # if this is a management request, fall thru to the command handler
      $r->handler('perl-script');
      $r->push_handlers(PerlHandler => \&command_handler);
      return OK;
   }
   my $prefix = $r->dir_config('ZopeProxyPrefix') || '/';
   my ($object, $parms) = split(/\?/, $uri, 2);
   return DECLINED if $object !~ /^$prefix/;
   my $base = $r->dir_config('ZopeProxyBaseURL');
   return DECLINED if !$base;
   $base =~ s/\/$// if $uri =~ /^\//;
   my $path = $r->dir_config('ZopeProxyPath') || DEFAULT_ZOPEPROXY_PATH;
   $path = $r->server_root_relative($path) if substr($path,0,1) ne '/';
   $path .= '/' if $path !~ /\/$/;
   if (!defined($objects{$base})) {
      initializeBase $path, $base, $r;
   }
   return proxyRequest($base, $uri, $r, 'ZopeProxyBase not correctly set')
      if !$objects{$base};
   return proxyRequest($base, $uri, $r)
      if !$objects{$base}->{$object};
   return proxyRequest($base, $uri, $r)
      if $r->method() ne 'GET';
   if ($anon{$base}->{$object}) {
      return proxyRequest($base, $uri, $r)
         if scalar($r->header_in('Authorization'));
   }
   my $hash = MD5->hexhash($object);
   return proxyRequest($base, $uri, $r, "$path.$hash not available")
      if ! -d $path.$hash;
   my $signature = getSignature($base, $object, $parms, $r);
   if (-r $path.$hash.$signature) {
      # already cached, process directly
      my $mtime = (stat($path.$hash.$signature))[9];
      if ($mtime + $timeout{$base}->{$object} >= time) {
         # still current, deliver from cache
         $r->filename($path.$hash.$signature);
         $r->handler('perl-script');
         $r->push_handlers(PerlTypeHandler => \&type_handler);
         return OK;
      } else {
         # not current any more, rebuild
         $r->handler('perl-script');
         $r->push_handlers(PerlHandler => \&proxy_handler);
         return OK;
      }
   } else {
      # not already cache, process by perl routine
      $r->handler('perl-script');
      $r->push_handlers(PerlHandler => \&proxy_handler);
      return OK;
   }
}

sub command_handler {
   # handle commands
   my $r = shift;
   my $uri = $r->uri;
   if ($uri =~ /^\/manage_ZopeProxyRemoteAction\/(.+?)(\/.*)$/) {
      my $command = $1;
      my ($object, $parms) = split(/\?/, $2, 2);
      $object = $object;
      my $ip = $r->connection()->remote_ip();
      if (!defined($allowedip)) {
         $allowedip = {};
         for $ip (split(/\s+/, $r->dir_config('ZopeProxyManagerIP'))) {
	    $allowedip->{$ip} = 1;
	 }
      }
      if ($allowedip && $allowedip->{$ip}) {
         my $base = $r->dir_config('ZopeProxyBaseURL');
         return DECLINED if !$base;
         $base =~ s/\/$// if $uri =~ /^\//;
         my $path = $r->dir_config('ZopeProxyPath') || DEFAULT_ZOPEPROXY_PATH;
         $path = $r->server_root_relative($path) if substr($path,0,1) ne '/';
         $path .= '/' if $path !~ /\/$/;
         if (!defined($objects{$base})) {
	    initializeBase $path, $base, $r;
	 }
         if ($command eq 'PURGE') {
	    if ($objects{$base}->{$object}) {
	       my $hash = MD5->hexhash($object);
	       if (opendir DIR, $path.$hash) {
	          my @files = readdir(DIR);
	          closedir DIR;
		  foreach my $file (@files) {
		     next if $file eq '.';
		     next if $file eq '..';
		     unlink $path.$hash.'/'.$file;
		  }
		  $r->content_type('text/plain');
		  $r->send_http_header;
		  $r->print('Object purged');
		  return OK;
	       } else {
	          $r->log_reason("Can't read directory $path$hash");
	          return BAD_REQUEST;
	       }
	    } else {
	       $r->log_reason("$object is not a cached object");
	       return BAD_REQUEST;
	    }
         } elsif ($command eq 'FLUSH') {
	    if (!defined($objects{$base})) {
	       $r->log_reason("ZopeProxyBase is not correctly set/read");
	       return BAD_REQUEST;
	    }
	    for my $object (keys %{$objects{$base}}) {
	       my $hash = MD5->hexhash($object);
	       if (opendir DIR, $path.$hash) {
	          my @files = readdir(DIR);
	          closedir DIR;
		  foreach my $file (@files) {
		     next if $file eq '.';
		     next if $file eq '..';
		     unlink $path.$hash.'/'.$file;
		  }
	       } else {
	          $r->log_reason("Can't read directory $path$hash");
	       }
	    }
	    $r->content_type('text/plain');
	    $r->send_http_header;
 	    $r->print('Cache Flushed');
	    return OK;
	 } elsif ($command eq 'RELOAD') {
	    initializeBase $path, $base, $r;
	    $r->content_type('text/plain');
	    $r->send_http_header;
	    $r->print('Config reloaded');
	    return OK;
	 } else {
	    $r->log_reason("Unknown command $command");
	    return BAD_REQUEST;
	 }
      } else {
         $r->log_reason("$ip has no access to manage_ZopeProxyRemoteAction");
         return FORBIDDEN;
      }
   } else {
      $r->log_reason("$uri has the wrong format");
      return BAD_REQUEST;
   }
}

sub type_handler {
   # just patch the content type
   my $r = shift;
   my $uri = $r->uri;
   my ($object, $parms) = split(/\?/, $uri);
   my $base = $r->dir_config('ZopeProxyBaseURL');
   return DECLINED if !$base;
   $r->content_type($objects{$base}->{$object});
}

sub proxy_handler {
   my $start = time();
   my $r = shift;
   my $base = $r->dir_config('ZopeProxyBaseURL');
   return DECLINED if !$base;
   my $path = $r->dir_config('ZopeProxyPath') || DEFAULT_ZOPEPROXY_PATH;
   $path = $r->server_root_relative($path) if substr($path,0,1) ne '/';
   $path .= '/' if $path !~ /\/$/;
   my $uri = $r->uri;
   $base =~ s/\/$// if $uri =~ /^\//;

   my $request = HTTP::Request->new($r->method, $base . $uri);
   $r->headers_in->do(sub {
      $request->header(@_);
   });
   if ($r->method eq 'POST') {
      my $len = $r->header_in('Content-length');
      my $buf;
      $r->read($buf, $len);
      $request->content($buf);
   }

   my $response = $UA->request($request);
   
   $r->content_type($response->header('Content-type'));
   $r->status($response->code);
   $r->status_line(join ' ', $response->code, $response->message);
   $response->scan(sub {
      $r->header_out(@_);
   });
   if ($r->header_only) {
      $r->send_http_header();
      return OK;
   }
   
   if ($response->is_success) {
      my ($object, $parms) = split(/\?/, $uri, 2);
      my $signature = getSignature($base, $object, $parms, $r);
      my $hash = MD5->hexhash($object);
      my $fh = Apache::File->new('>'.$path.$hash.$signature);
      if ($fh) {
         print $fh $response->content;
	 if ($response->header('Content-type') eq 'text/html') {
	    my $datum = scalar localtime;
	    print $fh "\n<!-- serverd by ZopeProxy cache date as of $datum -->\n";
	    $datum = scalar localtime(time+$timeout{$base}->{$object});
	    print $fh "<!-- Document $path$hash$signature valid until $datum -->\n";
	 }
      } else {
         $r->log_reason("Can't create $path$hash$signature");
      }
   }

   $r->content_type($response->content_type);
   $r->send_http_header;
   my $runningfor = (time() - $start);
   my $content = $response->content;
   $content .= "\n<!-- served by ZopeProxy in $runningfor seconds -->\n"
      if $response->content_type eq 'text/html';
   $r->print($content || $response->error_as_HTML);

   return OK;
}

sub redirect_ok {return undef;}

1;

__END__

